from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_darwin_mini_arm(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    #[np.pi / 2, -np.pi / 90, np.pi / 90]
    link_info = [[0, 0, 0],  # "theta" rotation in z
                 [0.08, 0, 0],  # "d" translation in z
                 [np.pi / 2, 0, 0],  # "alfa" rotation in x
                 [0.05, 0.18, 0.23],  # "a" translation in x
                 [0, 0, 0]]  # joint type

    scale = 1
    n = 3

    opacity = 1

    col_model = [[], [], []]
    # Create 3d objects

    Q01 = Utils.rotz(np.pi/2) * Utils.rotz(link_info[0][0]) * Utils.trn([0, 0, link_info[1][0]]) * Utils.rotx(link_info[2][0]) * Utils.trn(
        [link_info[3][0], 0, 0])
    Q02 = Utils.rotz(-np.pi / 90) * Utils.rotz(link_info[0][1]) * Utils.trn([0, 0, link_info[1][1]]) * Utils.rotx(link_info[2][1]) * Utils.trn(
        [link_info[3][1], 0, 0])
    Q03 = Utils.rotz(np.pi / 90) * Utils.rotz(link_info[0][2]) * Utils.trn([0, 0, link_info[1][2]]) * Utils.rotx(link_info[2][2]) * Utils.trn(
        [link_info[3][2], 0, 0])

    base_3d_obj = []

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/darwin_ext_01.obj',
            0.004,
            Utils.inv_htm(Q01)  * Utils.trn([-0.315, 0.435, -0.325]) * Utils.rotz(-3.14) * Utils.roty(-3.14 / 2),
            MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity, side="DoubleSide"))]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320.obj',
            0.004,
            Utils.inv_htm(Q01 * Q02) * Utils.trn([0.004, 0.01, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
            MeshMaterial(metalness=1, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                         opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPO.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02) * Utils.trn([0.004, 0.01, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/frm_ext_02.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02) * Utils.trn([0.004, 0.01, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide"))
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320.obj',
            0.004,
            Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.19, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
            MeshMaterial(metalness=1, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                         opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPO.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.19, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPU.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.013, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPD_2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.0136, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPO.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.roty(-np.pi / 2) * Utils.rotx(-np.pi / 2) * Utils.trn(
                    [-0.325, -0.385, -0.0115]),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPD_1.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.0136, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPL-2b2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.0136, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPL-2b2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([-0.044, 0.0136, -0.02]) * Utils.rotx(
                    -3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPL-2b2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.46, 0.502, 0.075]) * Utils.rotx(3.14 / 2) * Utils.rotz(
                    3.14 / 2) * Utils.trn([-0.044, 0.0136, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPL-2b2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0, 0, 0.083]) * Utils.roty(3.14) * Utils.trn(
                    [0.46, 0.502, 0.075]) * Utils.rotx(3.14 / 2) * Utils.rotz(
                    3.14 / 2) * Utils.trn([-0.044, 0.0136, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPD_1.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.035, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPD_3.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.004, 0.035, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPL-2b2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.007, 0.905, 0.043]) * Utils.rotx(3.14) * Utils.roty(
                    -3.14 / 2) * Utils.trn([0, 0, 0.083]) * Utils.roty(3.14) * Utils.trn(
                    [0.46, 0.502, 0.075]) * Utils.rotx(3.14 / 2) * Utils.rotz(
                    3.14 / 2) * Utils.trn([-0.044, 0.0136, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide")),
            Model3D(
                'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPL-2b2.obj',
                0.004,
                Utils.inv_htm(Q01 * Q02 * Q03) * Utils.trn([0.055, 0.905, 0.043]) * Utils.rotx(3.14) * Utils.roty(
                    -3.14 / 2) * Utils.trn([0, 0, 0.083]) * Utils.roty(3.14) * Utils.trn(
                    [0.46, 0.502, 0.075]) * Utils.rotx(3.14 / 2) * Utils.rotz(
                    3.14 / 2) * Utils.trn([-0.044, 0.0136, -0.02]) * Utils.rotx(-3.14 / 12) * Utils.trn(
                    [0, 0.2, 0]) * Utils.rotz(3.14) * Utils.trn([0.318, -0.2, -0.3]) * Utils.roty(-3.14 / 2),
                MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                             opacity=opacity, side="DoubleSide"))
        ]
    )

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    q0 = [0, 0, 0]

    # Create joint limits
    joint_limits = (np.pi / 180) * np.matrix([[-180, 180], [-180, 180], [-180, 180]])

    ###
    # Get all variables in the global scope
    # for i in range(12):
    #     try:
    #         matrix_str = str(eval('htm'+str(i+1)))
    #         matrix_str = matrix_str.strip()

    #         # Replace multiple spaces with a single space
    #         matrix_str = "\n".join(" ".join(row.split()) for row in matrix_str.split("\n"))

    #         # Split into rows, remove extra brackets
    #         rows = matrix_str.strip("[]").split("\n")

    #         # Format each row to insert commas between numbers
    #         formatted_rows = []
    #         for row in rows:
    #             row = row.strip()
    #             elements = row.strip("[]").split()
    #             formatted_row = "[" + ", ".join(elements) + "]"
    #             formatted_rows.append(formatted_row)

    #         # Join rows in a single line
    #         formatted_matrix = "[" + ", ".join(formatted_rows) + "]"
    #         print("htm"+str(i+1)+" = np.matrix("+formatted_matrix+")")
    #     except:
    #         print("")
    ###
    
    return base_3d_obj, links, np.identity(4), np.identity(4), q0, joint_limits



from utils import *
import numpy as np
from robot import *

def funF(r):
    f = np.matrix(r)
    for j in range(np.shape(r)[0]):
        f[j,0] = np.sign(r[j,0])*np.sqrt(np.abs(r[j,0]))
        
    return f

def _constrained_control(self, htm_tg, q=None, obstacles=[], htm=None, 
                              Kp =  2.0, eta_obs = 0.3, eta_auto = 0.3, eta_joint = 0.3, 
                              eps_to_obs = 0.003, h_to_obs = 0.003, 
                              eps_auto = 0.02,  h_auto = 0.05, 
                              d_safe_obs = 0.02, d_safe_auto = 0.002, d_safe_jl = (np.pi/180)*5,
                              eps_reg = 0.01):

    n = len(self.links)
    
    if str(type(obstacles))=="<class 'list'>":
        obstacles_list = obstacles
    else:
        obstacles_list = [obstacles]

    if q is None:
        q = self.q
    if htm is None:
        htm = self.htm

    q_min = self.joint_limit[:,0]
    q_max = self.joint_limit[:,1]
    
    #Get smooth distance structure from robot to point cloud
    A_obj = np.matrix(np.zeros((0,n)))
    b_obj_raw =np.matrix(np.zeros((0,1)))
    

    
    for obj in obstacles_list:    
        dr_obj =self.compute_dist(obj = obj, q = q, eps=eps_to_obs, h=h_to_obs)
        A_obj = np.matrix(np.vstack(  (A_obj, dr_obj.jac_dist_mat)  ))
        b_obj_raw = np.matrix(np.vstack(  (b_obj_raw, dr_obj.dist_vect)  ))

        
        
    b_obj = -eta_obs * (b_obj_raw-d_safe_obs)
    
    #Get smooth distance structure from robot to itself (autocollision)
    dr_auto = self.compute_dist_auto(q = q, eps=eps_auto, h=h_auto)    

    A_auto = dr_auto.jac_dist_mat
    b_auto_raw = dr_auto.dist_vect
    
    b_auto = -eta_auto * (b_auto_raw-d_safe_auto)
    
    #Assemble matrices for joint limit constraints
        
    A_joint = np.matrix(np.vstack(  (np.identity(7), -np.identity(7))  ))
    b_joint_raw = np.matrix(np.vstack(  (q-q_min , q_max - q) )) 
    
    b_joint = -eta_joint * (b_joint_raw-d_safe_jl)
    
    #Get task function data
    
    r, Jr = self.task_function(htm_tg, q=q)
    
    #Create the optimization problem
    A =    np.matrix(np.vstack( (A_obj, A_auto, A_joint) ) )
    b =    np.matrix(np.vstack( (b_obj, b_auto, b_joint) ) )
    H = 2*(Jr.transpose() * Jr + eps_reg *np.identity(7))
    f = Jr.transpose() * Kp * funF(r)
    
    #Solve QP and obtain joint velocity
    qdot = Utils.solve_qp(H, f, A, b)
    
    #Compute some statistics
    ep = np.linalg.norm(r[0:3,0])
    eox = np.arccos(1-r[3,0])
    eoy = np.arccos(1-r[4,0])
    eoz = np.arccos(1-r[5,0])
    eo = (180/np.pi)* max([eox,eoy,eoz])
    
    dobs = b_obj_raw.min()
    dauto = b_auto_raw.min()
    djoint = b_joint_raw.min()
    
 
    return qdot, ep, eo, dobs, dauto, djoint



from utils import *
from typing import Optional, Tuple, List

class Texture:
    """
  Texture to be applied in a Mesh Material.
  It is essentially a wrapper of threejs' 'Texture' class.
  See 'https://threejs.org/docs/#api/en/textures/Texture' for more details.

  Parameters
  ----------
  url : string
      The url that contains the texture.
      It must have one of the following formats: 'png', 'bmp', 'jpg', 'jpeg'.

  mapping : string
      Mirrors 'Texture.mapping' in threejs.
      How the image is applied to the object.
      A string  'UVMapping' is the default, where the U,V coordinates are used to apply the map.
      It can be 'UVMapping', 'CubeReflectionMapping', 'CubeRefractionMapping',
      'EquirectangularReflectionMapping', 'EquirectangularRefractionMapping',
      'CubeUVReflectionMapping' or 'CubeUVRefractionMapping'.
      (default: 'UVMapping').

  wrap_s : string
      Mirrors 'Texture.wrapS' in threejs.
      This defines how the texture is wrapped horizontally and corresponds to U in UV mapping.
      The default is 'ClampToEdgeWrapping', where the edge is clamped to the outer edge texels.
      It can be 'RepeatWrapping', 'ClampToEdgeWrapping' or 'MirroredRepeatWrapping'.
      (default: 'ClampToEdgeWrapping').

  wrap_t : string
      Mirrors 'Texture.wrapT' in threejs.
      This defines how the texture is wrapped vertically and corresponds to V in UV mapping.
      The default is 'ClampToEdgeWrapping', where the edge is clamped to the outer edge texels.
      It can be 'RepeatWrapping', 'ClampToEdgeWrapping' or 'MirroredRepeatWrapping'.
      (default: 'ClampToEdgeWrapping').

  mag_filter : string
      Mirrors 'Texture.magFilter' in threejs.
      How the texture is sampled when a texel covers more than one pixel.
      The default is 'LinearFilter', which takes the four closest texels and bilinearly interpolates among them.
      It can be 'NearestFilter' or 'LinearFilter'.
      (default: 'LinearFilter').

  min_filter : string
      Mirrors 'Texture.minFilter' in threejs.
      How the texture is sampled when a texel covers less than one pixel.
      The default is 'LinearMipmapLinearFilter', which uses mipmapping and a trilinear filter.
      It can be 'NearestFilter', 'NearestMipmapNearestFilter', 'NearestMipmapLinearFilter', 'LinearFilter',
      'LinearMipmapNearestFilter' or  'LinearMipmapLinearFilter'.
      (default: 'LinearMipmapLinearFilter').

  offset : 2d array
      Mirrors 'Texture.offset' in threejs.
      How much a single repetition of the texture is offset from the beginning, in each direction U and V.
      Typical range is 0.0 to 1.0.
      (default: [0,0]).

  repeat : 2d array
      Mirrors 'Texture.repeat' in threejs.
      How many times the texture is repeated across the surface, in each direction U and V.
      If repeat is set greater than 1 in either direction, the corresponding Wrap parameter should also be set to
      'RepeatWrapping' or 'MirroredRepeatWrapping' to achieve the desired tiling effect.
      Setting different repeat values for textures is restricted in the same way like .offset.
      (default: [1,1]).

  rotation : float
      How much the texture is rotated around the center point, in radians. Positive values are counter-clockwise.
      (default: 0).

  center : 2d array
      The point around which rotation occurs. A value of (0.5, 0.5) corresponds to the center of the texture.
      Default is [0, 0], the lower left.
      (default: [0, 0]).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def url(self) -> str:
        """The address of the texture."""
        return self._url

    @property
    def mapping(self) -> str:
        """The method for mapping of texture."""
        return self._mapping

    @property
    def wrap_s(self) -> str:
        """The method for wrapping in the U coordinate."""
        return self._wrap_s

    @property
    def wrap_t(self) -> str:
        """The method for wrapping in the V coordinate."""
        return self._wrap_t

    @property
    def mag_filter(self) -> str:
        """The method for the magnification filter."""
        return self._mag_filter

    @property
    def min_filter(self) -> str:
        """The method for the mignification filter."""
        return self._min_filter

    @property
    def offset(self) -> List[float]:
        """The offset in the texture."""
        return self._offset

    @property
    def repeat(self) -> List[float]:
        """The repeat pattern."""
        return self._repeat

    @property
    def rotation(self) -> float:
        """The texture rotation."""
        return self._rotation

    @property
    def center(self) -> List[float]:
        """The shifting for the texture center."""
        return self._center

    #######################################
    # Constructor
    #######################################

    def __init__(self, url: str, mapping: str ="UVMapping", wrap_s: str ="ClampToEdgeWrapping", 
                 wrap_t: str ="ClampToEdgeWrapping", mag_filter: str ="LinearFilter", 
                 min_filter: str ="LinearMipmapLinearFilter", offset: List[float] =[0, 0], 
                 repeat: List[float]=[1, 1], rotation: float =0, center: List[float] =[0, 0])->"Texture":

        image_types = ["png", "bmp", "jpg", "jpeg"]
        mapping_list = ['UVMapping', 'CubeReflectionMapping', 'CubeRefractionMapping',
                        'EquirectangularReflectionMapping', 'EquirectangularRefractionMapping',
                        'CubeUVReflectionMapping', 'CubeUVRefractionMapping']
        wrapping_mode_list = ['RepeatWrapping', 'ClampToEdgeWrapping', 'MirroredRepeatWrapping']
        mag_filter_list = ['NearestFilter', 'LinearFilter']
        min_filter_list = ['NearestFilter', 'NearestMipmapNearestFilter', 'NearestMipmapLinearFilter', 'LinearFilter',
                           'LinearMipmapNearestFilter', 'LinearMipmapLinearFilter']

        error = Utils.is_url_available(url, image_types)
        if not (error == "ok!"):
            raise Exception("The parameter 'url' " + error)

        if not (str(type(mapping)) == "<class 'str'>" and (mapping in mapping_list)):
            raise Exception(
                "The parameter 'mapping' should be one of the following strings: " + str(mapping_list) + ".")

        if not (str(type(wrap_s)) == "<class 'str'>" and (wrap_s in wrapping_mode_list)):
            raise Exception(
                "The parameter 'wrap_s' should be one of the following strings: " + str(wrapping_mode_list) + ".")

        if not (str(type(wrap_t)) == "<class 'str'>" and (wrap_t in wrapping_mode_list)):
            raise Exception(
                "The parameter 'wrap_t' should be one of the following strings: " + str(wrapping_mode_list) + ".")

        if not (str(type(mag_filter)) == "<class 'str'>" and (mag_filter in mag_filter_list)):
            raise Exception(
                "The parameter 'mag_filter' should be one of the following strings: " + str(mag_filter_list) + ".")

        if not (str(type(min_filter)) == "<class 'str'>" and (min_filter in min_filter_list)):
            raise Exception(
                "The parameter 'min_filter_list' should be one of the following strings: " + str(min_filter_list) + ".")

        if not Utils.is_a_vector(offset, 2):
            raise Exception("The parameter 'offset' should be 2D vector.")

        if not Utils.is_a_vector(repeat, 2):
            raise Exception("The parameter 'repeat' should be 2D vector.")

        if not Utils.is_a_number(rotation):
            raise Exception("The parameter 'rotation' should be a float.")

        if not Utils.is_a_vector(center, 2):
            raise Exception("The parameter 'center' should be 2D vector.")

        self._url = url
        self._mapping = mapping
        self._wrap_s = wrap_s
        self._wrap_t = wrap_t
        self._mag_filter = mag_filter
        self._min_filter = min_filter
        self._offset = offset
        self._repeat = repeat
        self._rotation = rotation
        self._center = center

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Texture: '" + self.url + "'"

        return string

    #######################################
    # Methods
    #######################################

    def gen_code(self, name):

        string = "const texture_" + name + " = new TextureLoader().load('" + self.url + "');\n"
        string += "texture_" + name + ".wrapS = " + str(self.wrap_s) + ";\n"
        string += "texture_" + name + ".wrapT = " + str(self.wrap_t) + ";\n"
        string += "texture_" + name + ".magFilter = " + str(self.mag_filter) + ";\n"
        string += "texture_" + name + ".minFilter = " + str(self.min_filter) + ";\n"
        string += "texture_" + name + ".rotation = " + str(self.rotation) + ";\n"
        string += "texture_" + name + ".offset.set(" + str(self.offset[0]) + "," + str(self.offset[1]) + ");\n"
        string += "texture_" + name + ".repeat.set(" + str(self.repeat[0]) + "," + str(self.repeat[1]) + ");\n"
        string += "texture_" + name + ".center.set(" + str(self.center[0]) + "," + str(self.center[1]) + ");\n\n"

        return string



from utils import *

from graphics.meshmaterial import *
from graphics.glbmeshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_kuka_lbr_iiwa(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    link_info = [[0, 0, 0, 0, 0, 0, 0],
                 [0.36, 0, 0.42, 0, 0.4, 0, 0.152],  # "d" translation in z
                 [-3.14 / 2, 3.14 / 2, 3.14 / 2, -3.14 / 2, -3.14 / 2, 3.14 / 2, 0],  # "alfa" rotation in x
                 [0, 0, 0, 0, 0, 0, 0],  # "a" translation in x 0.25
                 [0, 0, 0, 0, 0, 0, 0]]

    ############
    n = 7
    scale = 1
    dy=-0.02

    # Collision model
    col_model = [[], [], [], [], [], [], []]

    col_model[0].append(Cylinder(htm=np.matrix([[1,0,0,0] , [0,0,-1,0.19-dy], [0,1,0,0], [0,0,0,1]]) ,
                                 name=name + "_C0_0", radius=0.10, height=0.21, color="red", opacity=0.3))
    col_model[0].append(Ball(htm=Utils.trn([0.01, 0.01-dy, 0]),
                             name=name + "_C0_1", radius=0.13, color="red", opacity=0.3))

    col_model[1].append(Ball(htm=Utils.trn([0, 0.03, 0.04+dy]),
                             name=name + "_C1_0", radius=0.12, color="green", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([0, 0.02, 0.2+dy+0.0125]),
                                 name=name + "_C1_1", radius=0.09, height=0.145, color="green", opacity=0.3))

    col_model[2].append(Cylinder(htm=np.matrix([[1,0,0,0] , [0,0,1,-0.08+dy], [0,-1,0,0], [0,0,0,1]]),
                                 name=name + "_C2_0", radius=0.08, height=0.1, color="blue", opacity=0.3))
    col_model[2].append(Ball(htm=Utils.trn([0, -0.01+dy, -0.01]),
                             name=name + "_C2_1", radius=0.115, color="blue", opacity=0.3))

    col_model[3].append(Ball(htm=Utils.trn([0, -0.035, 0.063+2*dy]),
                             name=name + "_C3_0", radius=0.09, color="orange", opacity=0.3))
    col_model[3].append(Cylinder(htm=Utils.trn([0, -0.01, 0.195+2*dy]),
                                 name=name + "_C3_1", radius=0.09, height=0.125, color="orange", opacity=0.3))

    col_model[4].append(Cylinder(htm=np.matrix([[1,0,0,0] , [0,0,-1,0.088-2*dy], [0,1,0,0], [0,0,0,1]]),
                                 name=name + "_C4_0", radius=0.075, height=0.11, color="magenta", opacity=0.3))
    col_model[4].append(Box(htm=np.matrix([[1,0,0,0] , [0,0,-1,-0.01-2*dy], [0,1,0,-0.08], [0,0,0,1]]),
                            name=name + "_C4_1", width=0.12, depth=0.04, height=0.20, color="magenta", opacity=0.3))

    col_model[5].append(Cylinder(htm=Utils.trn([0, 0, 0.063+2*dy]),
                                 name=name + "_C5_0", radius=0.075, height=0.18, color="cyan", opacity=0.3))


    # Create 3d objects
    htm1 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, -1., 0.], [0., 1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm2 = np.matrix([[1., 0., -0.0016, 0.], [0., -1., 0., 0.36], [-0.0016, 0., -1., 0.], [0., 0., 0., 1.]])
    htm3 = np.matrix([[1., 0., -0.0016, 0.], [-0.0016, 0., -1., 0.], [0., 1., 0., 0.], [0., 0., 0., 1.]])
    htm4 = np.matrix([[1., 0., -0.0016, 0.], [0., 1., 0., -0.42], [0.0016, 0., 1., 0.], [0., 0., 0., 1.]])
    htm5 = np.matrix([[1., 0., -0.0016, 0.], [-0.0016, 0., -1., 0.], [0., 1., 0., 0.], [0., 0., 0., 1.]])
    htm6 = np.matrix([[1., 0., 0., 0.], [0., -1., 0., 0.4], [0., 0., -1., 0.], [0., 0., 0., 1.]])
    htm7 = np.matrix([[1., 0., 0., 0.], [0., 0., -1., 0.], [0., 1., 0., 0.], [0., 0., 0., 1.]])
    htm8 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, -1., 0.], [0., 1., 0.0008, -0.17], [0., 0., 0., 1.]])
    
    default_material_0 = MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5],
                     color=color, opacity=opacity)
    default_material_1 = MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#707070",
                         opacity=opacity)
    
    base_3d_obj = [Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_base_link.glb',
        scale,
        htm1 ,
        GLBMeshMaterial(opacity=opacity) if color == '' else default_material_0)]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_link_1.glb',
            scale,
            htm2,
            GLBMeshMaterial(opacity=opacity) if color == '' else default_material_0)
        ]
    )


    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_link_2.glb',
            scale,
            htm3,
            GLBMeshMaterial(opacity=opacity) if color == '' else default_material_0)
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_link_3.glb',
            scale,
            htm4,
            GLBMeshMaterial(opacity=opacity) if color == '' else default_material_0)
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_link_4.glb',
            scale,
            htm5,
            GLBMeshMaterial(opacity=opacity) if color == '' else default_material_0)
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_link_5.glb',
            scale,
            htm6 ,
            GLBMeshMaterial(opacity=opacity) if color == '' else default_material_0)
        ]
    )
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_link_6.glb',
            scale,
            htm7,
            GLBMeshMaterial(opacity=opacity) if color == '' else default_material_0)
        ]
    )
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/KukaLBRIIWA/iiwa_link_7.glb',
            scale,
            htm8,
            GLBMeshMaterial(opacity=opacity) if color == '' else default_material_1)
        ]
    )


    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration

    q0 = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-170,170],[-120,120],[-170,170],[-120,120],[-170,170],[-120,120],[-175,175]])
  
    return base_3d_obj, links, np.identity(4), Utils.trn([0,0,-0.04]), q0, joint_limits



import uaibot as ub
import numpy as np
import requests
from io import BytesIO
import matplotlib.pyplot as plt
import sys


def _constrained_control_demo_1():
    # Get the point cloud data
    url = "https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/PointCloud/data_wall_with_hole.npy"
    wallpoints = np.load(BytesIO(requests.get(url).content))
    pc = ub.PointCloud(points = wallpoints, size=0.02, color='cyan')

    #Create the other objects and set up the simulation
    point_link = ub.Ball(color='red',radius=0.02)
    point_obj = ub.Ball(color='blue',radius=0.02)
    robot = ub.Robot.create_franka_emika_3()
    htm_tg = ub.Utils.trn([0.45,0.31,0.7])*ub.Utils.rotx(-np.pi/2)
    frame_tg = ub.Frame(htm_tg, size=0.1)
    sim = ub.Simulation.create_sim_grid([pc, robot, point_link, point_obj, frame_tg])
    sim.set_parameters(width=500,height=500,load_screen_color='#191919',background_color='#191919')

    #Define F function for control
    def funF(r):
        f = np.matrix(r)
        for j in range(np.shape(r)[0]):
            f[j,0] = np.sign(r[j,0])*np.sqrt(np.abs(r[j,0]))
            
        return f

    # for link in robot.links:
    #     for obj in link.col_objects:
    #         sim.add(obj[0])
            
    #Set some parameters
    dt = 0.02
    eta = 0.3
    Kp = 2.0
    Tmax = 35
    q_min = robot.joint_limit[:,0]
    q_max = robot.joint_limit[:,1]

    eps_to_obs = 0.003
    h_to_obs = 0.003
    eps_auto = 0.02
    h_auto = 0.05
    d_safe_obs = 0.02
    d_safe_auto = 0.002
    d_safe_jl = (np.pi/180)*5
    tolp=1.0 #1mm error tolerance
    tolo=1.0 #1deg error tolerance
    eps_reg = 0.01

    #Main loop
    t = 0

    robot.add_ani_frame(0,[0.0, 0.0, 0.0, -np.pi*4/180-2*d_safe_jl, 0.0, 2*d_safe_jl, 0.0])


    hist_data=[]
    cont = True

    while cont:
        
        #Get current configuration
        qr = robot.q
        
        #Get smooth distance structure from robot to point cloud
        dr_obj =robot.compute_dist(obj = pc, q = qr, eps=eps_to_obs, h=h_to_obs)
        
        A_obj = dr_obj.jac_dist_mat
        b_obj = dr_obj.dist_vect-d_safe_obs
        
        #Get smooth distance structure from robot to itself (autocollision)
        dr_auto =robot.compute_dist_auto(q = qr, eps=eps_auto, h=h_auto)    

        A_auto = dr_auto.jac_dist_mat
        b_auto = dr_auto.dist_vect-d_safe_auto
        
        #Assemble matrices for joint limit constraints
            
        A_joint = np.matrix(np.vstack(  (np.identity(7), -np.identity(7))  ))
        b_joint = np.matrix(np.vstack(  (qr-(q_min+d_safe_jl) , (q_max-d_safe_jl) - qr)  )) 
        
        #Get task function data
        
        r, Jr = robot.task_function(htm_tg, q=qr)
        
        #Create the optimization problem
        A =    np.matrix(np.vstack( (A_obj, A_auto, A_joint) ) )
        b =    -eta * np.matrix(np.vstack( (b_obj, b_auto, b_joint) ) )
        H = 2*(Jr.transpose() * Jr + eps_reg *np.identity(7))
        f = Jr.transpose() * Kp * funF(r)
        
        #Solve QP and obtain joint velocity
        
        qdot = ub.Utils.solve_qp(H, f, A, b)

        #Integrate joint velocity
        qr += qdot*dt
        
        #Compute some statics
        epx = abs(round(1000*r[0,0]))
        epy = abs(round(1000*r[1,0]))
        epz = abs(round(1000*r[2,0]))
        eox = round((180/np.pi)*np.arccos(1-r[3,0]),2)
        eoy = round((180/np.pi)*np.arccos(1-r[4,0]),2)
        eoz = round((180/np.pi)*np.arccos(1-r[5,0]),2)
        
        #Compute true distance to show
        dr_obj_0 =robot.compute_dist(obj = pc, q = qr)
        dr_auto_0 =robot.compute_dist_auto(q = qr)  
        
        dmin_obj = 1000*dr_obj_0.get_closest_item().distance
        dmin_auto = 1000*dr_auto_0.get_closest_item().distance
        
        str_msg = "\rt = "+str(round(t,2))+" s, epx = "+str(epx)+" mm, epy = "+str(epy)+" mm, epz = "+str(epz)+" mm, "
        str_msg += "eox = "+str(epx)+" deg, eoy = "+str(epy)+" deg, eoz = "+str(epz)+" deg     "
        
        prox_joints = (180/np.pi) * min((qr-q_min).min(), (q_max - qr).min())
        
        
        hist_data.append([t, epx, epy, epz, eox, eoy, eoz, dmin_obj, dmin_auto, prox_joints])

        
        sys.stdout.write(str_msg)
        sys.stdout.flush()
        
        #Update animation frames and time
        robot.add_ani_frame(t,qr)
        robot.update_col_object(t)
        point_link.add_ani_frame(t, htm = ub.Utils.trn(dr_obj.get_closest_item().point_link))
        point_obj.add_ani_frame(t, htm = ub.Utils.trn(dr_obj.get_closest_item().point_object ))            
        t+=dt
        
        #Check condition
        cont = t < Tmax and ((eox > tolo) or (eoy > tolo) or (eoz > tolo) or (epx > tolp) or (epy > tolp) or (epz > tolp))
        

    # Convert list to NumPy array for easier slicing
    data = np.array(hist_data)

    # Extract columns
    t = data[:, 0]  # Time
    epx, epy, epz = data[:, 1], data[:, 2], data[:, 3]
    eox, eoy, eoz = data[:, 4], data[:, 5], data[:, 6]
    dmin_obj, dmin_auto = data[:, 7], data[:, 8]
    prox_joints = data[:, 9]

    # Create a figure with subplots
    fig, axs = plt.subplots(4, 1, figsize=(8, 12), sharex=True)

    # Plot epx, epy, epz vs. time
    axs[0].plot(t, epx, label='epx')
    axs[0].plot(t, epy, label='epy')
    axs[0].plot(t, epz, label='epz')
    axs[0].set_ylabel("Position error (mm)")
    axs[0].legend()
    axs[0].grid(True)

    # Plot eox, eoy, eoz vs. time
    axs[1].plot(t, eox, label='eox')
    axs[1].plot(t, eoy, label='eoy')
    axs[1].plot(t, eoz, label='eoz')
    axs[1].set_ylabel("Orientation error (mm)")
    axs[1].legend()
    axs[1].grid(True)

    # Plot dmin_obj, dmin_auto vs. time
    axs[2].plot(t, dmin_obj, label='to obj')
    axs[2].plot(t, dmin_auto, label='auto')
    axs[2].set_ylabel("Collision distance (mm)")
    axs[2].legend()
    axs[2].grid(True)

    # Plot prox_joints vs. time
    axs[3].plot(t, prox_joints, label='Joint limit margin')
    axs[3].set_xlabel("Time (t)")
    axs[3].set_ylabel("Joint limit margin (deg)")
    axs[3].grid(True)

    # Show the plots
    plt.tight_layout()
    plt.show()
    
    return sim

# sim.save("/home/vinicius/Desktop/uaibot_devel/uaibot_files/testing","intro_3")


from utils import *
import numpy as np

def _check_free_config(self, q=None, htm=None, obstacles=[],
                              check_joint=True, check_auto=True,
                              tol=0.0005, dist_tol=0.005, no_iter_max=20, mode='auto'):
    
    n = len(self.links)

    if q is None:
        q = self.q
    if htm is None:
        htm = self.htm

    # Error handling
    if mode not in ['python','c++','auto']:
        raise Exception("The parameter 'mode' should be 'python,'c++', or 'auto'.")

    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not str(type(obstacles)) == "<class 'list'>":
        raise Exception("The parameter 'list' should be a list of simple objects.")

    for obs in obstacles:
        if not Utils.is_a_metric_object(obs):
            raise Exception("The parameter 'list' should be a list of metric objects : " + str(Utils.IS_METRIC) + ".")

    if not str(type(check_joint)) == "<class 'bool'>":
        raise Exception("The parameter 'check_joint' should be a boolean.")

    if not str(type(check_auto)) == "<class 'bool'>":
        raise Exception("The parameter 'check_auto' should be a boolean.")

    if not Utils.is_a_number(tol) or tol <= 0:
        raise Exception("The parameter 'tol' should be a positive number.")

    if not Utils.is_a_number(dist_tol) or dist_tol <= 0:
        raise Exception("The parameter 'dist_tol' should be a positive number.")

    if not Utils.is_a_natural_number(no_iter_max) or no_iter_max <= 0:
        raise Exception("The parameter 'no_iter_max' should be a positive natural number.")

    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
        raise Exception("c++ mode is set, but .so file was not loaded!")
    # end error handling

    if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        return _check_free_config_python(self, Utils.cvt(q), htm, obstacles, check_joint, check_auto, tol, dist_tol, no_iter_max)
    else:
        obstacles_cpp=[]
        for obs in obstacles:
            if Utils.get_uaibot_type(obs) == 'uaibot.CPP_GeometricPrimitives':
                obstacles_cpp.append(obs)
            else:
                obstacles_cpp.append(Utils.obj_to_cpp(obs))
                
        cfcres = self.cpp_robot.check_free_configuration(Utils.cvt(q), htm, obstacles_cpp, check_joint, check_auto, tol, dist_tol, no_iter_max)
        return cfcres.isfree, cfcres.message, cfcres.info


        
def _check_free_config_python(self, q=None, htm=None, obstacles=[],
                              check_joint=True, check_auto=True,
                              tol=0.0005, dist_tol=0.005, no_iter_max=20):
    
    n = len(self.links)

    #_, mth_dh = self.jac_geo(q, "dh", htm)
    mth_dh = self.fkm(q, "dh", htm, mode='python')

    col_object_copy = []

    # Update all collision objects of all links
    for i in range(n):
        col_object_copy.append([])
        for j in range(len(self.links[i].col_objects)):
            temp_copy = self.links[i].col_objects[j][0].copy()
            htmd = self.links[i].col_objects[j][1]
            temp_copy.set_ani_frame(mth_dh[i][:, :] * htmd)
            col_object_copy[i].append(temp_copy)


    #Create the list of all possible auto collision check
    list_auto_col_check = []
    for i in range(n):
        for j in range(i+2,n):
            for isub in range(len(self.links[i].col_objects)):
                for jsub in range(len(self.links[j].col_objects)):
                    list_auto_col_check.append([i,isub,j,jsub])

    #Create the list of all possible collisions
    list_col_check = []
    for i in range(n):
        for isub in range(len(self.links[i].col_objects)):
            for j in range(len(obstacles)):
                list_col_check.append([i,isub,j])

    #Check if there is any violation
    collided = False
    message = "Ok!"
    info = []

    # Check joint limits:
    if check_joint:
        k = 0
        while not collided and k < n:
            collided = q[k] < self.joint_limit[k, 0]
            if collided:
                message = "Joint number " + str(k) + " is below minimum limit."
                info = [0, k]

            k += 1

        k = 0
        while not collided and k < n:
            collided = q[k] > self.joint_limit[k, 1]
            if collided:
                message = "Joint number " + str(k) + " is above maximum limit."
                info = [1, k]

            k += 1

    # Check collision with the obstacles
    k = 0
    while not collided and k < len(list_col_check):

        i = list_col_check[k][0]
        isub = list_col_check[k][1]
        j = list_col_check[k][2]

        if Utils.compute_aabbdist(obstacles[j], col_object_copy[i][isub]) == 0:
            _, _, d, _ = Utils.compute_dist(obstacles[j], col_object_copy[i][isub] \
                                                     , None, tol, no_iter_max, mode='python')
            if d < dist_tol:
                collided = True
                message = "Collision between link " + str(i) + " (col object " + str(isub) + ") and obstacle "+str(j)+"."
                info = [2, i, isub, j]

        k += 1


    # Check auto collision:
    if check_auto:
        k = 0
        while not collided and k < len(list_auto_col_check):

            i = list_auto_col_check[k][0]
            isub = list_auto_col_check[k][1]
            j = list_auto_col_check[k][2]
            jsub = list_auto_col_check[k][3]

            if Utils.compute_aabbdist(col_object_copy[i][isub], col_object_copy[j][jsub]) == 0:

                _, _, d, _ = Utils.compute_dist(col_object_copy[i][isub], col_object_copy[j][jsub] \
                                                         , None, tol, no_iter_max, mode='python')
                
                if d < dist_tol:
                    collided = True
                    message = "Collision between link " + str(i) + " (col object " + str(isub) + ") and link " + str(
                        j) + " " \
                             "(col object " + str(jsub) + ")."
                    info = [3, i, isub, j, jsub]

            k += 1




    return (not collided), message, info



from utils_test import *
import uaibot as ub
import time
import sys
import os

print("Running Object to Object Distance Test")

t = 0

projball1 = ub.Ball(radius=0.02,color='blue')
projball2 = ub.Ball(radius=0.02,color='red')

sim = ub.Simulation.create_sim_grid([projball1, projball2])

t = 0

sum_time = 0
max_time = 0
cross_error_max = 0
N=100



for i in range(N):
    
    cont=True
        
    while cont:
        obj1 = generate_rand_object(color='red')
        obj2 = generate_rand_object(color='blue')
        
        type1 = ub.Utils.get_uaibot_type(obj1)
        type2 = ub.Utils.get_uaibot_type(obj2)
        cont =  type1 == 'uaibot.PointCloud' and type2 == 'uaibot.PointCloud'

        if not cont:
            start = time.time() 
            p1, p2, d, _ = ub.Utils.compute_dist(obj1, obj2)
            end = time.time() 
            cont = d < 0.2
            
            #Cross checking
            if not cont:
                error1 = np.linalg.norm(obj1.projection(p2)[0]-p1)
                error2 = np.linalg.norm(obj2.projection(p1)[0]-p2)
                cross_error_max = np.max([cross_error_max,error1, error2])
                
                # if error1 > 1e-3 or error2 > 1e-3:
                #     print("Warning, check "+str(i)+", "+type1+", "+type2+", error1 = "+str(round(error1,2))+", error2 = "+str(round(error2,2)))

    
    elapsed = end - start
    sum_time+=elapsed
    max_time=max(max_time,elapsed) 
    
    projball1.add_ani_frame(i,ub.Utils.trn(p1))
    projball2.add_ani_frame(i,ub.Utils.trn(p2))   

    sim.add(obj1)
    sim.add(obj2)
                         
    if type1 != 'uaibot.PointCloud':
        htm1 = np.matrix(obj1.htm)
        
    if type2 != 'uaibot.PointCloud':    
        htm2 = np.matrix(obj2.htm)
    
    if type1 == 'uaibot.PointCloud':
        obj1.add_ani_frame(0,0,0)
        obj1.add_ani_frame(i,0,np.shape(obj1.points)[1])
        obj1.add_ani_frame(i+1,0,0)   
    else:
        obj1.add_ani_frame(0,htm=ub.Utils.trn([1000,1000,1000]))
        obj1.add_ani_frame(i,htm=htm1)
        obj1.add_ani_frame(i+1,htm=ub.Utils.trn([1000,1000,1000]))        

    if type2 == 'uaibot.PointCloud':
        obj2.add_ani_frame(0,0,0)
        obj2.add_ani_frame(i,0,np.shape(obj2.points)[1])
        obj2.add_ani_frame(i+1,0,0) 
    else:
        obj2.add_ani_frame(0,htm=ub.Utils.trn([1000,1000,1000]))
        obj2.add_ani_frame(i,htm=htm2)
        obj2.add_ani_frame(i+1,htm=ub.Utils.trn([1000,1000,1000]))
        


    sys.stdout.write("\rCompleted: "+str(round(100*i/N))+"%")
    sys.stdout.flush()

             
print("\nAverage time took to compute distance: "+str(round(1000*sum_time/N,2))+" ms")
print("Max time took to compute distance: "+str(round(1000*max_time,2))+" ms")
print("Maximum cross error: "+str(round(1000*cross_error_max,3))+" mm")

print("Test was a success!")   

        

        

        
    
    

current_folder = os.path.dirname(os.path.abspath(__file__))           
sim.save(current_folder,"test_o2o")



import numpy as np
from utils import *

from graphics.meshmaterial import *
from graphics.mtlmeshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_franka_emika_3(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color) and not color=='':
        raise Exception("The parameter 'color' should be a HTML-compatible color or empty.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")



    link_info = [[0.00000, 0.0000, 0.0000, 0.00000, 0.00000, 0.0000, 0.0000],
                 [0.33300, 0.0000, 0.3160, 0.00000, 0.38400, 0.0000, 0.1070+0.07],  # "d" translation in z
                 [-1.5708, 1.5708, 1.5708, -1.5708, 1.57080, 1.5708, 0.0000],  # "alfa" rotation in x
                 [0.00000, 0.0000, 0.0825, -0.0825, 0.00000, 0.0880, 0.0000],  # "a" translation in x 0.25
                 [0, 0, 0, 0, 0, 0, 0]]

    n = 7
    scale = 1


    # Collision model
    col_model = [[], [], [], [], [], [], []]

    col_model[0].append(Cylinder(htm= Utils.trn([0,0,0]),
                                 name=name + "_C0_0", radius=0.08, height=0.28, color="red", opacity=0.3))

    col_model[1].append(Cylinder(htm= Utils.trn([0,0,0.18]),
                                 name=name + "_C1_0", radius=0.07, height=0.25, color="blue", opacity=0.3))
        
    col_model[2].append(Cylinder(htm= Utils.trn([0,0,0]),
                                 name=name + "_C2_0", radius=0.07, height=0.24, color="green", opacity=0.3))

    col_model[3].append(Cylinder(htm= Utils.trn([0,0,0.13]),
                                 name=name + "_C3_0", radius=0.07, height=0.20, color="yellow", opacity=0.3))

    A = np.matrix([ [1,0,0,0], [0,0,1,-0.383], [0,-1,0,0],[0,0,0,1]])

    col_model[4].append(Box(htm= A * Utils.trn([0,0.09,0.28]) * Utils.rotx(-0.24),
                                 name=name + "_C4_0", width=0.08, depth=0.05, height=0.23, color="magenta", opacity=0.3))

    col_model[4].append(Cylinder(htm= Utils.trn([0,0,-0.03]),
                                 name=name + "_C4_1", radius=0.06, height=0.22, color="magenta", opacity=0.3))

    col_model[5].append(Cylinder(htm= Utils.trn([0,0,0.02]),
                                 name=name + "_C5_0", radius=0.06, height=0.21, color="cyan", opacity=0.3))

    col_model[6].append(Box(htm= Utils.trn([0,0,0.04-0.09]), 
                            name=name + "_C6_0", width=0.05 , depth=0.21, height=0.10, color="red", opacity=0.3))

    col_model[6].append(Box(htm= Utils.trn([0.03, 0.03, -0.025-0.07]) * Utils.rotz(np.pi / 4), 
                            name=name + "_C6_1", width=0.1000 , depth=0.09, height=0.0340, color="red", opacity=0.3))


    # Create 3d objects
    htm1 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, -1., 0.], [0., 1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm2 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, 1., 0.], [0., -1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm3 = np.matrix([[1., 0., 0., -0.0825], [0., 0.0008, 1., 0.], [0., -1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm4 = np.matrix([[1., 0., 0., 0.0825], [0., 0.0008, -1., 0.], [0., 1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm5 = np.matrix([[1., 0., 0., 0.], [0., 0.0008, 1., 0.], [0., -1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm6 = np.matrix([[1., 0., 0., -0.088], [0., 0.0008, 1., 0.], [0., -1., 0.0008, 0.], [0., 0., 0., 1.]])
    htm7 = np.matrix([[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., -0.177], [0., 0., 0., 1.]])
    htm8 = np.matrix([[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., -0.067], [0., 0., 0., 1.]])  

    link_3d_obj = []
    
    default_material = MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5],color=color, opacity=opacity)

    base_3d_obj = [Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link0.obj',
        scale,
        np.identity(4),
        MTLMeshMaterial(url = 'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link0.mtl', 
                        opacity=opacity) if color=='' else default_material) ]
    
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link1.obj',
            scale,
            htm1,
            MTLMeshMaterial(url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link1.mtl',
                         opacity=opacity) if color=='' else default_material)
        ]
    )
        

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link2.obj',
            scale,
            htm2,
            MTLMeshMaterial(url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link2.mtl',
                         opacity=opacity) if color=='' else default_material)
        ]
    )
    

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link3.obj',
            scale,
            htm3, 
            MTLMeshMaterial(url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link3.mtl',
                         opacity=opacity) if color=='' else default_material)
        ]
    )

 
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link4.obj',
            scale,
            htm4, 
            MTLMeshMaterial(url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link4.mtl',
                         opacity=opacity) if color=='' else default_material)
        ]
    )
            
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link5.obj',
            scale,
            htm5, 
            MTLMeshMaterial(url = 'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link5.mtl',
                         opacity=opacity) if color=='' else default_material)
        ]
    )
 

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link6.obj',
            scale,
            htm6, 
            MTLMeshMaterial(url = 'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link6.mtl',
                         opacity=opacity) if color=='' else default_material)
        ]
    )


    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link7.obj',
            scale,
            htm7, 
            MTLMeshMaterial(url = 'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link7.mtl',
                         opacity=opacity) if color=='' else default_material),
        Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/hand.obj',
            scale,
            htm8,  
            MTLMeshMaterial(url = 'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/hand.mtl',
                            opacity=opacity) if color=='' else default_material)
        ]
    )
    
    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    q0 = [0.0, 0.0, 0.0, -np.pi*4/180, 0.0, 0.0, 0.0]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-166,166],[-101,101],[-166,166],[-176,-4],[-166,166],[-1,215],[-166,166]])
    
    return base_3d_obj, links, np.identity(4), np.identity(4), q0, joint_limits


import numpy as np
from utils import *
import uaibot_cpp_bind as ub_cpp



def _to_cpp(robot):

    n = len(robot.links)

    cpp_robot = ub_cpp.CPP_Manipulator(n)

    cpp_robot.set_htm_extra(robot.htm_base_0, robot.htm_n_eef)

    for i in range(n):
        cpp_robot.set_joint_param(i, robot.links[i].theta, robot.links[i].d, robot.links[i].alpha, robot.links[i].a, int(robot.links[i].joint_type), robot.joint_limit[i,0], robot.joint_limit[i,1])
        for j in range(len(robot.links[i].col_objects)):
            prim = robot.links[i].col_objects[j][0]
            htm = robot.links[i].col_objects[j][1]
            cpp_prim = Utils.obj_to_cpp(prim, htm)            
            cpp_robot.add_geo_prim(i, cpp_prim)
        
    
    return cpp_robot










import os, sys

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

import numpy as np
from utils import *
from graphics.meshmaterial import *
from simobjects.group import *
from simobjects.frame import *

from ._set_ani_frame import _set_ani_frame
from ._add_ani_frame import _add_ani_frame



from ._ikm import _ikm
from ._fkm import _fkm
from ._jac_geo import _jac_geo
from ._jac_jac_geo import _jac_jac_geo
from ._jac_ana import _jac_ana

from ._dyn_model import _dyn_model

from ._vector_field_rn import _vector_field_rn
from ._task_function import _task_function
from ._vector_field_SE3 import _vector_field_SE3, _DELTA as _DELTA_SE3, _DS as _DS_SE3


from ._gen_code import _gen_code
from ._update_col_object import _update_col_object
from ._add_col_object import _add_col_object
from ._attach_object import _attach_object
from ._detach_object import _detach_object

from ._compute_dist import _compute_dist
from ._compute_dist_auto import _compute_dist_auto
from ._check_free_config import _check_free_config

from ._create_kuka_kr5 import _create_kuka_kr5
from ._create_epson_t6 import _create_epson_t6
from ._create_staubli_tx60 import _create_staubli_tx60
from ._create_kuka_lbr_iiwa import _create_kuka_lbr_iiwa
from ._create_abb_crb import _create_abb_crb
from ._create_darwin_mini import _create_darwin_mini
from ._create_franka_emika_3 import _create_franka_emika_3
from ._create_davinci import _create_davinci
from ._create_magician_e6 import _create_magician_e6
from ._create_kinova_gen3 import _create_kinova_gen3
from ._create_jaco import _create_jaco

from .links import *

from ._constrained_control import _constrained_control
from uaibot.simulation.simulation import *

from ._dist_struct_robot_obj import *
from ._dist_struct_robot_auto import *

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from uaibot.robot import DistStructRobotObj
    from uaibot.robot import DistStructRobotAuto
    from uaibot.robot import Link
    from uaibot.simulation import Simulation

from numpy.typing import ArrayLike, NDArray

import os
if os.environ['CPP_SO_FOUND']=="1":
    import uaibot_cpp_bind as ub_cpp
    from ._to_cpp import _to_cpp

from uaibot.utils.types import HTMatrix, Matrix, Vector, MetricObject, GroupableObject
from typing import Optional, Tuple, List


class Robot:
    """
  A class that contains a robot object in UAIBot.

  Parameters
  ----------

  name : string
      The robot name.
      (default: 'genRobot').

  htm_base_0 : 4x4 numpy matrix
      The transformation between the robot's base and the first DH frame.
      (default: 4x4 identity matrix).

  htm_n_eef : 4x4 numpy matrix
      The transformation between the last DH and the end-effector.
      (default: 4x4 identity matrix).

  list_object_3d_base : list of 'uaibot.Model3D' objects
      The list of 3d models of the base of the robot.
      If set to None, there is no base 3d object.
      (default: None).

  links : A list of 'uaibot.Link' objects
      The list of link objects.

  q0 : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
      The robot initial configuration.
      (default: zero vector).

  eef_frame_visible : boolean
      Set if the end-effector frame is visible.
      (default: True).

  joint_limit : n x 2 numpy array or None
      A n x 2 numpy array containing the joint limits, either in rad or meters.
      If set to 'None', use very large joint limits.
      (default: None).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def q(self) -> Vector:
        """The current joint configuration."""
        return np.matrix(self._q)

    @property
    def q0(self) -> Vector:
        """The default joint configuration."""
        return np.matrix(self._q0)

    @property
    def htm(self) -> HTMatrix:
        """
        The current base configuration in scenario coordinates.
        A 4x4 homogeneous matrix written is scenario coordinates.
        """
        return np.matrix(self._htm)

    @property
    def htm_base_0(self) -> HTMatrix:
        """
        The constant homogeneous transformation between the HTM of the base and
        the HTM of the first Denavit-Hartenberg frame.
        """
        return np.matrix(self._htm_base_0)

    @property
    def htm_n_eef(self) -> HTMatrix:
        """
        The constant homogeneous transformation between the HTM of the last
        Denavit-Hartenberg frame and the end-effector
        """
        return np.matrix(self._htm_n_eef)

    @property
    def links(self) -> List["Link"]:
        """Data structures containing the links of the robot."""
        return self._links

    @property
    def attached_objects(self) -> List[GroupableObject]:
        """Data structures containing the objects attached into the robot."""
        return self._attached_objects

    @property
    def name(self) -> str:
        """Name of the object."""
        return self._name

    @property
    def list_object_3d_base(self) -> List[Model3D]:
        """The list of 3d objects that form the base."""
        return self._list_object_3d_base

    @property
    def eef_frame_visible(self) -> bool:
        """If the frame attached to the end effector is visible"""
        return self._eef_frame_visible

    @property
    def joint_limit(self) -> np.matrix:
        """A n x 2 numpy array containing the joint limits, either in rad or meters"""
        return self._joint_limit

    @property
    def cpp_robot(self):
        """Used in the c++ interface"""
        return self._cpp_robot
    
    #######################################
    # Constructor
    #######################################

    def __init__(self, name: str, links: List["Link"], list_base_3d_obj : Optional[HTMatrix] = None, 
                 htm: HTMatrix =np.identity(4), htm_base_0: HTMatrix = np.identity(4),
                 htm_n_eef: HTMatrix = np.identity(4), q0 : Optional[Vector] = None, 
                 eef_frame_visible: bool =True, joint_limits: Optional[np.matrix] = None) -> "Robot":
        # Error handling

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_matrix(htm_base_0, 4, 4):
            raise Exception("The parameter 'htm_base_0' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_matrix(htm_n_eef, 4, 4):
            raise Exception("The parameter 'htm_n_eef' should be a 4x4 homogeneous transformation matrix.")

        if not (str(type(links)) == "<class 'list'>"):
            raise Exception("The parameter 'links' should be a list of 'uaibot.Link' objects.")
        else:
            for link in links:
                if not Utils.get_uaibot_type(link) == "uaibot.Link":
                    raise Exception("The parameter 'links' should be a list of 'uaibot.Link' objects.")

        n = len(links)
        
        if name=="":
            name="var_robot_id_"+str(id(self))
            
        if not (q0 is None):
            self._q0 = np.matrix(q0).reshape((n,1))
        else:
            self._q0 = np.matrix(np.zeros((n,1)))

        if not (str(type(list_base_3d_obj)) == "<class 'list'>" or (list_base_3d_obj is None)):
            raise Exception("The parameter 'list_base_3d_obj' should be a list of 'uaibot.Model3D' objects.")
        else:
            for i in range(len(list_base_3d_obj)):
                if not (Utils.get_uaibot_type(list_base_3d_obj[i]) == "uaibot.Model3D"):
                    raise Exception(
                        "The parameter 'list_base_3d_obj' should be a list of 'uaibot.Model3D' objects.")

        if not Utils.is_a_vector(self._q0, n):
            raise Exception("The parameter 'q0' should be a " + str(n) + " dimensional vector.")

        if not str(type(eef_frame_visible)) == "<class 'bool'>":
            raise Exception("The parameter 'eef_frame_visible' must be a boolean.")

        if not joint_limits is None:
            self._joint_limit = joint_limits
        else:
            self._joint_limit = np.block([-10 * np.ones((n, 1)) , 10 * np.ones((n, 1))])

        if not Utils.is_a_matrix(self._joint_limit, n, 2):
            raise Exception("The parameter 'joint_limits' should be a " + str(n) + " x 2 numpy array.")

        for i in range(n):
            if self._joint_limit[i, 0]>self._joint_limit[i, 0]:
                raise Exception(
                    "In the parameter 'joint_limits', the minimum value (first column) must be smaller or equal than the maximum value (second column)")

        # end error handling

        self._frames = []
        self._list_object_3d_base = list_base_3d_obj
        self._htm = np.matrix(htm)
        self._name = name
        self._attached_objects = []
        self._links = links
        self._htm_base_0 = htm_base_0
        self._htm_n_eef = htm_n_eef
        self._eef_frame_visible = eef_frame_visible

        
        if eef_frame_visible:
            self._eef_frame = Frame(size=0.1)
            
        self._max_time = 0

        #Create the cpp robot
        if os.environ['CPP_SO_FOUND']=="1":
            self._cpp_robot = _to_cpp(self)
            
        # Set initial total configuration
        self.set_ani_frame(self._q0, self._htm)



    #######################################
    # Std. Print
    #######################################

    def __repr__(self):
        n = len(self.links)

        string = "Robot with name '" + self.name + "': \n\n"
        string += " Number of joints: " + str(n) + "\n"
        string += " Joint types: "

        for i in range(n):
            string += "R" if self._links[i].joint_type == 0 else "P"

        string += "\n"
        string += " Current configuration: " + str([round(num[0,0], 3) for num in self.q]) + "\n"
        string += " Current base HTM: \n" + str(self.htm) + "\n"
        string += " Current end-effector HTM: \n" + str(self.fkm())
        return string

    #######################################
    # Methods for configuration changing
    #######################################

    def add_ani_frame(self, time: float, q: Optional[Vector]=None, htm: Optional[HTMatrix]=None, 
                      enforce_joint_limits: bool = False) -> None:
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration.
    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current HTM).
    enforce_joint_limits: boolean
        If True and some q violates the joint limits (seen in the attribute 'joint_limit'), the
        respective q is clamped. Note that it DOES NOT issue any warning for this. So, be aware.
        (default: False).

    Returns
    -------
    None
    """
        return _add_ani_frame(self, time, q, htm, enforce_joint_limits)

    def set_ani_frame(self, q: Optional[Vector]=None, htm: Optional[HTMatrix]=None, 
                      enforce_joint_limits: bool = False) -> None:
        """
    Reset object's animation queue and add a single configuration to the 
    object's animation queue.

    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration.
        (default: the current joint configuration (robot.q) for the manipulator, q0).
    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current HTM).
    enforce_joint_limits: boolean
        If True and some q violates the joint limits (seen in the attribute 'joint_limit'), the
        respective q is clamped. Note that it DOES NOT issue any warning for this. So, be aware.
        (default: False).
    Returns
    -------
    None
    """
        return _set_ani_frame(self, q, htm, enforce_joint_limits)

    #######################################
    # Methods for kinematics model
    #######################################

    def fkm(self, q: Optional[Vector] = None, axis: str ='eef', htm: Optional[HTMatrix]=None, 
            mode: str ='auto') -> List[HTMatrix]:
        """
    Compute the forward kinematics for an axis at a given joint and base
    configuration. Everything is written in the scenario coordinates.

    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).
    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector;
        'dh': for all Denavit-Hartenberg axis;
        'com': for all center-of-mass axis.
        (default: 'eef').    
    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current HTM).
    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').

    Returns
    -------
    htm_fk : list of 4x4 numpy matrices
        For axis='eef', returns a single htm. For the other cases, return
        n htms as a list of 4x4 numpy matrices.
    """
        return _fkm(self, q, axis, htm, mode)


    def ikm(self, htm_target: Optional[HTMatrix]=None, htm_tg: Optional[HTMatrix]=None, 
            htm: Optional[HTMatrix]=None, q0: Optional[Vector]=None, p_tol: float=0.001, 
            a_tol: float =5, no_iter_max: int =200, ignore_orientation: bool = False, 
            no_tries = 40, check_joint: bool = True, check_auto: bool = True, 
            obstacles: List[MetricObject]=[], mode: str ='auto') -> Vector:
        """
    Try to solve the inverse kinematic problem for the end-effector, given a
    desired homogeneous transformation matrix. It returns the manipulator
    configuration.

    Uses an iterative algorithm.

    The algorithm can fail, throwing an Exception when it happens.

    Parameters
    ----------
    htm_tg : 4x4 numpy matrix
        The target end-effector HTM, written in scenario coordinates.
    htm_target : 4x4 numpy matrix
        The target end-effector HTM, written in scenario coordinates.
        This is deprecated for compability, use 'htm_tg' instead. You can leave this as 'None'.        
    htm : 4x4 numpy matrix
        The pose of the basis of the manipulator.
        (default: 'None' (the current base htm))
    q0 : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        Initial guess for the algorithm for the joint configuration.
        (default: a random joint configuration).
    p_tol : positive float
        The accepted error for the end-effector position, in meters.
        (default: 0.001 m).    
    a_tol : positive float
        The accepted error for the end-effector orientation, in degrees.
        (default: 5 degrees). 
    no_iter_max : positive int
        The maximum number of iterations for the algorithm.
        (default: 200 iterations).
    ignore_orientation: boolean
        If True, the orientation part of the HTM is ignored. Task is position-only.
        (default: False)
    no_tries: positive int.
        How many times the algorithm tries to find a solution.
        (default: 40)
    check_joint: boolen
        If True, consider the joint limits as well.
        (default: True)
    check_auto: boolen
        If True, consider the auto collision of the robot.
        (default: True)
    obstacles: list of 'MetricObject' objects
        List of objects as obstacles to consider.
        (default: Empty list)
    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').

    Returns
    -------
    q : n x 1 numpy matrix
        The configuration that solves the IK problem.
    """
    
    
        # Backward compatibility shim
        if htm_tg is None and htm_target is not None:
            htm_tg = htm_target
            import warnings
            warnings.warn(
                "'htm_target' is deprecated, use 'htm_tg' instead.",
                DeprecationWarning
            )
        elif htm_tg is None:
            raise ValueError("Missing required argument: 'htm_tg'")    
        
    
        return _ikm(self, htm_tg, htm, q0, p_tol, a_tol, no_iter_max, ignore_orientation, 
                    no_tries, check_joint, check_auto, obstacles, mode)

    def jac_geo(self, q: Optional[Vector] = None, axis: str ='eef', htm : Optional[HTMatrix]=None, 
                mode: str ='auto') -> Tuple[np.matrix,HTMatrix]:
        """
    Compute the geometric Jacobian for an axis at a given joint and base
    configuration. Also returns the forward kinematics as a by-product.
    Everything is written in the scenario coordinates.

    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration 
        (default: the current  joint configuration (robot.q) for the manipulator).
    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector;
        'dh': for all Denavit-Hartenberg axis;
        'com': for all center-of-mass axis.
        (default: 'eef').    
    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current htm).
    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').

    Returns
    -------
    jac_geo : 6 x n or list of 6 x n numpy matrices
        For axis='eef', returns a single 6xn Jacobian. For the other cases, 
        return n Jacobians as a list of n 6xn numpy matrices.

    htm_out : 4 x 4 or list of 4 x 4 numpy matrices
        For axis='eef', returns a single htm as a 4x4 matrix. For the other cases, return
        n htms as  a list of n 4x4 numpy matrices.
    """
        return _jac_geo(self, q, axis, htm, mode)

    def jac_ana(self, q: Optional[Vector]=None, htm: Optional[HTMatrix]=None) -> Tuple[Matrix,HTMatrix,Vector]:
        """
    Compute the analytic Jacobian for the end-effector. The Euler angle
    convention is zyx. Also returns the end-effector htm and Euler angles
    as a by-product.

    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration
        (default: the current  joint configuration (robot.q) for the manipulator).
    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current htm).

    Returns
    -------
    jac_ana : 6 x n numpy matrix
        The analytic Jacobian.

    htm_eef : 4 x 4 numpy matrix
        The end-effector htm.

    phi : 3x1 numpy matrix
        The euler angles in the z (alpha), y (beta) and x (gamma) convention,
        as a column vector.
    """
        return _jac_ana(self, q, htm)

    def jac_jac_geo(self, q: Optional[Vector] =None, axis: str ='eef', 
                    htm: Optional[HTMatrix] = None)-> List[np.matrix]:
        """
    Compute the Jacobians of the columns of the geometric Jacobian in the joint variable 'q'.
    This can be either to the end-effector frames (axis='eef'), to the Denavit-Hartenberg (DH) frames
    (axis='dh') or to the center-of-mass (COM) frames (axis='com').

    If axis ='dh' or 'com':

    If the robot has n links, jj_geo = robot.jac_jac_geo(q, htm) is a list of lists, in which
    jj_geo[i][j] is a 6 x n matrix that represents the Jacobian of the j-th column of the geometric
    Jacobian matrix of the i-th DH or COM frame.

    jj_geo[i][j] for j>i is not computed, because the j-th column of the geometric Jacobian matrix of the
    i-th DH or COM frame is always the 6 x n zero matrix, regardless of the 'q' and 'htm' chosen.

    jj_geo[i][j] could be alternatively computed numerically. For example, for axis='dh', by defining the function of q
    f = lambda q_var: np.matrix(robot.jac_geo(q = q_var, htm = htm, axis = 'dh')[0][i][j])
    and then computing the numerical Jacobian as Utils.jac(f,q).
    However, this function  is faster and has greater numerical accuracy, since it is computed analytically
    instead of numerically.

    If axis='eef', this is the same as computing axis='dh' but throwing away all but the last list away.


    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration
        (default: the current  joint configuration (robot.q) for the manipulator).

    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector;
        'dh': for all Denavit-Hartenberg axis;
        'com': for all center-of-mass axis.
        (default: 'eef').

    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current htm).

    Returns
    -------
    jj_geo : list of lists of 6 x n numpy arrays (if axis='dh' or 'com') or a list of 6 x n numpy arrays (if axis=='eef')
        The Jacobian of the j-th column of the geometric Jacobian matrix of the i-th Denavit-Hartenberg frame.

    """
        return _jac_jac_geo(self, q, axis, htm)

    #######################################
    # Methods for control
    #######################################
    @staticmethod
    def vector_field(q, curve: List[Vector], alpha: float =1, 
                     const_vel: float =1, is_closed: bool = True, gamma: float=10, mode: str ='auto') -> Tuple[np.matrix,float,int]:
        """
    Computes the vector field presented in 
    
    "Adriano M. C. Rezende; Vinicius M. Goncalves; Luciano C. A. Pimenta: 
    Constructive Time-Varying Vector Fields for Robot Navigation 
    IEEE Transactions on Robotics (2021)". 
    
    The vector field has constant velocity and use the function 
    G(p) = (2/pi)*atan(alpha*sqrt(Dist(p))).
    
    in which "Dist(p)" is the Euclidean distance to the curve and alpha>0 a parameter.
    
    If the curve is not closed, the circulation component is modulated by:
    
    R(p) = min(gamma*(1.0 - s(p)),1.0)
    
    in which s(p) in [0,1] is the fraction of the curve travelled at the point p*(p) 
    (i.e., the closest point to the curve to point p) and gamma>0.
    

    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The configuration in which the vector field should be computed.

    curve : a list of nD vectors (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        Curve, described as sampled points. 
 
    alpha : positive float
        Controls the vector field behaviour. Greater alpha's imply more 
        robustness to the vector field, but increased velocity and acceleration
        behaviours. Used in G(u) = (2/pi)*atan(alpha*u)
        (default: 1).

    const_vel : positive float
        The constant velocity of the vector field. The signal of this number 
        controls the direction of rotation 
        (default: 1).
        
    is_closed: bool
        If the curve is closed or not.
        (default: True)
        
    gamma: positive float
        The parameter of the function that sends the circulation 
        component to zero when close to the end of the curve. 
        Only active when it 'is_closed' is True.
        (default: 0.5)

    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').

    Returns
    -------
    qdot : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The velocity generated by the vector field.

    dist : float
        Distance to the curve.

    index : integer
        The index of the closest point to the curve.
    """

        return _vector_field_rn(q, curve, alpha, const_vel, is_closed, gamma, mode)
 
    def task_function(self, htm_target: Optional[HTMatrix]=None, 
                      htm_tg: Optional[HTMatrix]=None, q: Optional[None]=None, 
                      htm: Optional[Vector] = None, mode: str = 'auto') -> Tuple[np.matrix,np.matrix]:
        """
    Computes the 6-dimensional task function for end-effector pose control,  
    given a joint configuration, a base configuration and the desired pose 
    'htm_tg'.

    The first three entries are position error, and the three last entries are
    orientation error.

    Everything is written in scenario coordinates. 

    Also returns the Jacobian of this function.

    Parameters
    ----------
    htm_tg : 4x4 numpy array 
        The target end-effector pose. 

    htm_target : 4x4 numpy matrix
        The target end-effector HTM, written in scenario coordinates.
        This is deprecated for compability, use 'htm_tg' instead. You can leave this as 'None'.  
         
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).

    htm : 4x4 numpy array 
        The robot base's configuration.
        (default: the same as the current htm).

    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').

    Returns
    -------
    r : 6 x 1 numpy matrix
        The task function.

    jac_r : 6 x n numpy matrix
        The respective task Jacobian.
    """
    
        # Backward compatibility shim
        if htm_tg is None and htm_target is not None:
            htm_tg = htm_target
            import warnings
            warnings.warn(
                "'htm_target' is deprecated, use 'htm_tg' instead.",
                DeprecationWarning
            )
        elif htm_tg is None:
            raise ValueError("Missing required argument: 'htm_tg'")      
    
    
        return _task_function(self, htm_tg, q, htm, mode)

    def constrained_control(self, htm_tg: HTMatrix, q: Optional[Vector] = None, obstacles: List[MetricObject]=[], 
                            htm: Optional[HTMatrix] = None, Kp: float =  2.0, eta_obs: float = 0.3, eta_auto: float = 0.3, eta_joint: float = 0.3, 
                              eps_to_obs: float = 0.003, h_to_obs: float = 0.003, eps_auto: float = 0.02,  h_auto: float = 0.05, 
                              d_safe_obs: float = 0.02, d_safe_auto: float = 0.002, d_safe_jl: float = (np.pi/180)*5,
                              eps_reg: float = 0.01) -> Tuple[np.matrix, float, float, float, float]:
        

        
        return _constrained_control(self, htm_tg, q, obstacles, htm, Kp, eta_obs, eta_auto, eta_joint, 
                              eps_to_obs, h_to_obs, eps_auto,  h_auto, d_safe_obs, d_safe_auto, d_safe_jl,eps_reg)
    
    @staticmethod
    def vector_field_SE3(
        state: HTMatrix,
        curve: ArrayLike | List[HTMatrix],
        kt1: float = 1.0,
        kt2: float = 1.0,
        kt3: float = 1.0,
        kn1: float = 1.0,
        kn2: float = 1.0,
        curve_derivative: ArrayLike | List[HTMatrix] = [],
        delta: float = _DELTA_SE3,
        ds: float = _DS_SE3,
        mode: str = 'auto'
        ) -> Tuple[NDArray, float, int]:
        r"""Computes the vector field in SE(3) presented in the thesis
        'Felipe B. A. Pessoa, Luciano C. A. Pimenta, Vinicius M. Goncalves: 
        Constructive Vector Fields for Path Following in Matrix Lie Groups,
        Master's Thesis, Universidade Federal de Minas Gerais, 2025'. 
    
        Parameters
        ----------
        state: 4x4 numpy array
            The current state of the robot, as a homogeneous transformation
            matrix.
        curve : nx4x4 numpy array or nx4x4 nested list
            Curve, described as sampled points. Each one of the n rows should 
            contain a 4x4 homogeneous transformation matrix, written in world
            coordinates.
        kt1 : float
            A gain parameter for the tangent component of the vector field.
            (default: 1.0).
        kt2 : float
            A gain parameter for the normal component of the vector field.
            (default: 1.0).
        kt3 : float
            A gain parameter for the binormal component of the vector field.
            (default: 1.0).
        kn1 : float
            A gain parameter for the normal component of the vector field.
            (default: 1.0).
        kn2 : float
            A gain parameter for the binormal component of the vector field.
            (default: 1.0).
        curve_derivative : nx4x4 numpy array or nx4x4 nested list
            The derivative of the curve, as a list of 4x4 homogeneous
            transformation matrices, written in world coordinates.
            If not provided, it is computed numerically.
            (default: empty list).
        delta : float
            Parameter that controls the precision of the normal compoennt
            approximation.
            (default: 1e-3).
        ds : float
            Parameter that controls the precision of the tangent component
            approximation. Ignored if curve_derivative is provided.
            (default: 1e-3).
        mode : string
            'c++' for the c++ implementation, 'python' for the python 
            implementation and 'auto' for automatic ('c++' is available, else 
            'python').
            (default: 'auto').

        Returns
        -------
        qdot : n x 1 numpy array
            The twist generated by the vector field.
        dist : float
            Distance to the curve.
        index : integer
            The index of the closest point to the curve.

        Notes
        -----
        The vector field (:math:`\Psi`) is defined as a combination of tangent
        (:math:`\xi_T`) and normal (:math:`\xi_N`) components [1]_:
        
        .. math:: \Psi = k_T \xi_T + k_N \xi_N,

        where :math:`k_T` and :math:`k_N` are gain parameters for the tangent
        and normal components, respectively. The normal component is computed
        to drive the following distance to zero

        .. math:: D(H) = \min_s \|\log(H^{-1} H_d(s))\|_F,

        where :math:`H` is the current state, :math:`H_d(s)` is a point on the
        curve `curve`.

        The gains are then defined as:

        .. math:: 
            k_N = k_{n1} \tanh(k_{n2} \sqrt{D(H)})\\
            k_T = k_{t1} (1 - k_{t2} \tanh(k_{t3} D(H))),

        where :math:`k_{n1}`, :math:`k_{n2}`, :math:`k_{t1}`, :math:`k_{t2}`,
        and :math:`k_{t3}` are the function parameters `kt1`, `kt2`, `kt3`,
        `kn1`, and `kn2`, respectively.

        References
        ----------
        .. [1] Felipe B. A. Pessoa, Luciano C. A. Pimenta, Vinicius M. Goncalves,
           "Constructive Vector Fields for Path Following in Matrix Lie Groups",
           Master's Thesis, Universidade Federal de Minas Gerais, 2025.
        """
        return _vector_field_SE3(state, curve, kt1, kt2, kt3, kn1, kn2, curve_derivative, delta, ds, mode)
        
        
    #######################################
    # Methods for simulation
    #######################################

    def gen_code(self):
        """Generate code for injection."""
        return _gen_code(self)

    def update_col_object(self, time: float, mode: str ='auto') -> None:
        """
        Update internally the objects that compose the collision model to the
        current configuration of the robot.
        """
        _update_col_object(self, time, mode)

    def add_col_object(self, sim: "Simulation") -> None:
        """
        Add the objects that compose the collision model to a simulation.

        Parameters
        ----------
        sim : 'Simulation' object
            'Simulation' object.
    """
        _add_col_object(self, sim)

    def attach_object(self, obj: GroupableObject) -> None:
        """
        Attach an object to the end-effector.
        The list of the type of objects that can be grouped can be seen in 'Utils.IS_GROUPABLE'.

        Parameters
        ----------
        obj : Object that is one of the types in 'Utils.IS_GROUPABLE'
            Object to be attached.
    """
        _attach_object(self, obj)

    def detach_object(self, obj: GroupableObject) -> None:
        """
        Detach an object (ball, box or cylinder) to the end-effector.

        Parameters
        ----------
        obj : object
            Object to be detached.
    """
        _detach_object(self, obj)

    def set_htm_to_eef(self, htm: HTMatrix) -> None:
        self._htm_n_eef = htm

    #######################################
    # Robot constructors
    #######################################

    @staticmethod
    def create_kuka_kr5(htm: HTMatrix = np.identity(4), name: str ='', color: str ="#df6c25", 
                        opacity: float =1, eef_frame_visible: bool=True) -> "Robot":
        """
    Create a Kuka KR5 R850, a six-degree of freedom manipulator.
    Thanks Sugi-Tjiu for the 3d model (see https://grabcad.com/library/kuka-kr-5-r850).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))
 
    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color.
        (default: '#df6c25')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_kuka_kr5(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_epson_t6(htm: HTMatrix = np.identity(4), name: str ='', color: str ="white", 
                        opacity: float =1, eef_frame_visible: bool = True)-> "Robot":
        """
    Create an Epson T6, a SCARA manipulator.
    Thanks KUA for the 3d model (see https://grabcad.com/library/epson-t6-scara-robot-1).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color.
        (default: 'white')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_epson_t6(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_staubli_tx60(htm: HTMatrix = np.identity(4), name: str = '', color: str ="#ff9b00", 
                            opacity: float =1, eef_frame_visible: bool = True) -> "Robot":
        """
    Create a Staubli TX60, a six degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/ros-industrial/staubli).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color.
        (default: '#ff9b00')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_staubli_tx60(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_kuka_lbr_iiwa(htm: HTMatrix = np.identity(4), name: str = '', color: str = "", 
                             opacity: float =1, eef_frame_visible: bool = True) -> "Robot":
        """
    Create a Kuka LBR IIWA 14 R820, a seven degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/ros-industrial/kuka_experimental).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color. 
        If set to '' (empty string), set colors from textures.
        (default: '')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_kuka_lbr_iiwa(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_franka_emika_3(htm: HTMatrix = np.identity(4), name: str = '', color : str = "", 
                              opacity: float = 1, eef_frame_visible: bool = True) -> "Robot":
        """
    Create a Franka Emika 3, a seven degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/BolunDai0216/FR3Env/tree/d5218531471cadafd395428f8c2033f6feeb3555/FR3Env/robots/meshes/visual).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color. 
        If set to '' (empty string), set colors from textures.
        (default: '')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_franka_emika_3(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_abb_crb(htm : HTMatrix = np.identity(4), name: str = '', color: str = "white", 
                       opacity: float = 1, eef_frame_visible: bool = True) -> "Robot":
        """
    Create a ABB CRB 15000, a six degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/ros-industrial/abb_experimental).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color.
        (default: 'white')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_abb_crb(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_magician_e6(htm: HTMatrix = np.identity(4), name: str = "", color: str = "#3e3f42", 
                           opacity: float = 1, eef_frame_visible: bool = True) -> "Robot":
        """
    Create a DOBOT Magician E6, a six degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/Dobot-Arm/TCP-IP-ROS-6AXis).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color.
        (default: 'white')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_magician_e6(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)
    
    @staticmethod
    def create_darwin_mini(htm: HTMatrix = np.identity(4), name: str = "", color: str = "#3e3f42", 
                           opacity: float = 1, eef_frame_visible: bool = True) -> GroupableObject:
        """
    Create an (oversized) Darwin Mini, a humanoid robot.
    Thanks to Alexandre Le Falher for the 3D model (https://grabcad.com/library/darwin-mini-1).

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: empty (automatic)).

    htm : color
        A HTML-compatible string representing the object color.
        (default: '#3e3f42').

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Group object
        The robot. It is composed of a group of six objects: the two arms and legs (members of 'Robot' class)
        and the chest and head (both 'RigidObject' class)


    """

        param_arm_left, param_arm_right, param_leg_left, param_leg_right, head, chest = _create_darwin_mini(htm, name,
                                                                                                            color,
                                                                                                            opacity)
        desl_z = htm * Utils.trn([0, 0, -0.18])
        robot_arm_left = Robot(name + "__arm_left", param_arm_left[1], param_arm_left[0],
                               desl_z * Utils.trn([0, 0.14, 1]) * Utils.rotx(-3.14 / 2), param_arm_left[2],
                               param_arm_left[3], [np.pi/2,0.3,0], eef_frame_visible,
                               param_arm_left[5])

        robot_arm_right = Robot(name + "__arm_right", param_arm_right[1], param_arm_right[0],
                                desl_z * Utils.rotz(3.14) * Utils.trn([0, 0.14, 1]) * Utils.rotx(-3.14 / 2),
                                param_arm_right[2],
                                param_arm_right[3], [np.pi/2,0.3,0], eef_frame_visible, param_arm_right[5])

        robot_leg_left = Robot(name + "__leg_left", param_leg_left[1], param_leg_left[0],
                               desl_z * Utils.trn([0, -0.1, 0.7]) * Utils.roty(np.pi / 2) * Utils.rotz(-np.pi / 2),
                               param_leg_left[2], param_leg_left[3], [np.pi/2, 0, 0, np.pi/2], eef_frame_visible, param_leg_left[5])

        robot_leg_right = Robot(name + "__leg_right", param_leg_right[1], param_leg_right[0],
                                desl_z * Utils.trn([0, 0.1, 0.7]) * Utils.roty(np.pi / 2) * Utils.rotz(-np.pi / 2),
                                param_leg_right[2], param_leg_right[3], [np.pi/2, 0, 0, np.pi/2], eef_frame_visible, param_leg_right[5])

        return Group([robot_arm_left, robot_arm_right, robot_leg_left, robot_leg_right, head, chest])

    @staticmethod
    def create_davinci(htm: HTMatrix = np.identity(4), name: str = "", color: str = "#3e3f42", 
                       opacity: float = 1, eef_frame_visible: bool = True) -> GroupableObject:
        """
        Create a da Vinci Si, a surgical robot.
        Thanks to Koray Okan for the 3D model (https://grabcad.com/library/da-vinci-surgical-robot-1/details).
        Created by Felipe Bartelt.
        Parameters
        ----------
        htm : 4x4 numpy matrix
            The initial base configuration for the robot.
            (default: np.identity(4))
        name : string
            The robot name.
            (default: empty (automatic)).
        htm : color
            A HTML-compatible string representing the object color.
            (default: '#3e3f42').
        opacity : positive float between 0 and 1
            The opacity of the robot. 1 = fully opaque and 0 = transparent.
            (default: 1)
        Returns
        -------
        robot : Group object
            The robot. It is composed of a group of six objects: the two arms and legs (members of 'Robot' class)
            and the chest and head (both 'RigidObject' class)
        """

        return _create_davinci(htm, name, color, opacity, eef_frame_visible)
    
    @staticmethod
    def create_kinova_gen3(
        htm: HTMatrix = np.eye(4),
        name: str = "",
        color: str = "#e6e1e1",
        opacity: float = 1.0,
        eef_frame_visible: bool = True
        ) -> "Robot":
        """Create a Kinova Gen3-7DoF, a seven degree of freedom 
        manipulator.

        Model from Kinova Resources (https://www.kinovarobotics.com/resources
        Gen3 CAD model (7DoF)). Manipulator parameters taken from the official 
        Kinova Gen3 documentation (GEN3 User Guide), available at
        https://www.kinovarobotics.com/resources.
        There is no gripper in this model, because it is not available.

        Parameters
        ----------
        htm : 4x4 numpy array or 4x4 nested list
            The initial base configuration for the robot.
            (default: np.identity(4))
        name : string
            The robot name.
            (default: empty (automatic)).
        color: string
            A HTML-compatible string representing the object color.
            (default: '#e6e1e1').
        opacity : positive float between 0 and 1
            The opacity of the robot. 1 = fully opaque and 0 = transparent.
            (default: 1)

        Returns
        -------
        robot : Robot object
            The robot.
        """

        return _create_kinova_gen3(htm, name, color, opacity, eef_frame_visible)
    
    @staticmethod
    def create_jaco(
        htm: HTMatrix = np.eye(4),
        name: str = "",
        color: str | list | None = None,
        opacity: float = 1.0,
        eef_frame_visible: bool = True
        ) -> "Robot":
        """Create a Kinova Jaco 2, a six degree of freedom manipulator.
        Model taken from Kinova resources 
        (https://www.kinovarobotics.com/resources?r=79302&s).
        Robot documentation available at 
        https://www.kinovarobotics.com/resources?r=339.
        The gripper is not yet controllable.

        Parameters
        ----------
        htm : 4x4 numpy array or 4x4 nested list
            The initial base configuration for the robot.
            (default: np.identity(4))
        name : string
            The robot name.
            (default: empty (automatic)).
        color : string or list or None
            A HTML-compatible string representing the object color or a list of
            HTML-compatible strings representing the object color. The list 
            should have three elements, representing the color of the links, 
            plastic rings, and the nails, respectively. If less than three 
            elements are provided, the remaining elements will be set to the
            last element of the list. If a single string is provided, it will
            be used for all three elements.
            If set to None, the default color scheme is used, which is a list
            of three strings: ["#3e3f42", "#919090", "#1d1d1f"].
        opacity : positive float between 0 and 1
            The opacity of the robot. 1 = fully opaque and 0 = transparent.
            (default: 1)

        Returns
        -------
        robot : Robot object
            The robot.
        """

        return _create_jaco(htm, name, color, opacity, eef_frame_visible)
    
    #######################################
    # Distance computation and collision
    #######################################

    def compute_dist(self, obj: MetricObject, q: Optional[Vector] = None, htm: Optional[HTMatrix]=None, 
                     old_dist_struct : Optional["DistStructRobotObj"] = None, tol: float = 0.0005, 
                     no_iter_max: int = 20, max_dist: float = np.inf, h: float = 0, eps: float = 0, 
                     mode: str = 'auto') -> "DistStructRobotObj":
        """
    Compute the  distance structure from each one of the robot's link to a
    'simple' external object (see Utils.IS_SIMPLE), given a joint and base
    configuration.

    This function can be faster if some distance computations are avoided.
    See the description of the parameter 'max_dist'.

    The distance is either Euclidean distance or differentiable.
    
    If h>0 or eps > 0, it computes the Euclidean distance and it uses GJK's algorithm.
    
    Else, it computes the differentiable distance through Generalized Alternating Projection (GAP).
    See the paper 'A Differentiable Distance Metric for Robotics Through Generalized Alternating Projection'.
    This only works in c++ mode, though.

    Parameters
    ----------
    obj : a simple object (see Utils.IS_SIMPLE)
        The external object for which the distance structure is going to be 
        computed, for each robot link.

    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).

    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current htm).

    old_dist_struct : 'DistStructRobotObj' object
        'DistStructRobotObj' obtained previously for the same robot and external object.
        Can be used to enhance the algorithm speed using the previous closest
        point as an initial guess.
        (default: None).

    tol : positive float
        Convergence criterion of GAP: it stops when ||a[k+1]-a[k]|| < tol.
        Only valid when h > 0 or eps > 0
        (default: 0.0005 m)      

    no_iter_max : positive int 
        Maximum number of iterations of GAP.
        Only valid when h > 0 or eps > 0
        (default: 20 iterations) 

    max_dist: positive float
        The algorithm uses an axis aligned bounding box (aabb) to avoid some distance computations.
        The algorithm skips a more precise distance computation if the distance between the aabb
        of the primitive objects composing the link and 'obj' is less than 'max_dist' (in meters).
        (default: infinite).

    h : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0) 

    eps : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0) 
                
    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').

    Returns
    -------
    dist_struct : 'DistStructRobotObj' object
        Distance struct for each one of the m objects that compose the robot's
        collision model. Contains a list of m 'DistStructLinkObj' objects.
    """

        return _compute_dist(self, obj, q, htm, old_dist_struct, tol, no_iter_max, max_dist, h, eps, mode)

    def compute_dist_auto(self, q: Optional[Vector] = None, old_dist_struct: Optional["DistStructRobotAuto"]=None, 
                          tol: float =0.0005, no_iter_max: int = 20, max_dist: float = np.inf, 
                          h: float = 0, eps: float = 0, mode: str = 'auto') -> "DistStructRobotAuto":
        """
    Compute the  distance structure from each one of the robot's links to itself
    (auto collision), given a joint and base configuration.

    This function consider only non-sequential links, since the collision between
    sequential links in the kinematic chain can be verified by checking joint limits.
    This saves times. This verification should be done elsewhere (by checking if the
    configuration is inside the joint limits).

    This function can be faster if some distance computations are avoided.
    See the description of the parameter 'max_dist'.

    The distance is either Euclidean distance or differentiable.
    
    If h>0 or eps > 0, it computes the Euclidean distance and it uses GJK's algorithm.
    
    Else, it computes the differentiable distance through Generalized Alternating Projection (GAP).
    See the paper 'A Differentiable Distance Metric for Robotics Through Generalized Alternating Projection'.
    This only works in c++ mode, though.

    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).

    old_dist_struct : 'DistStructRobotAuto' object
        'DistStructRobotAuto' obtained previously for the same robot.
        Can be used to enhance the algorithm speed using the previous closest
        point as an initial guess.
        (default: None).

    tol : positive float
        Convergence criterion of GAP: it stops when ||a[k+1]-a[k]|| < tol.
        Only valid when h > 0 or eps > 0.
        (default: 0.0005 m)      

    no_iter_max : positive int 
        Maximum number of iterations of GAP.
        Only valid when h > 0 or eps > 0
        (default: 20 iterations) 

    max_dist: positive float
        The algorithm uses an axis aligned bounding box (aabb) to avoid some distance computations.
        The algorithm skips a more precise distance computation if the distance between the aabb
        of the primitive objects composing the link and 'obj' is less than 'max_dist' (in meters).
        (default: infinite).

    h : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0) 

    eps : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0) 
        
    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').
        
    Returns
    -------
    dist_struct : 'DistStructRobotAuto' object
        Distance struct for each one of the m objects that compose the robot's
        collision model. Contains a list of m 'DistStructLinkLink' objects.
    """

        return _compute_dist_auto(self, q, old_dist_struct, tol, no_iter_max, max_dist, h, eps, mode)


        

    def check_free_config(self, q: Optional[Vector]=None, htm: Optional[HTMatrix]=None, 
                          obstacles: List[MetricObject]=[], check_joint: bool = True, check_auto: bool = True,
                          tol: float = 0.0005, dist_tol: float = 0.005, no_iter_max: int = 20, 
                          mode: str = 'auto') -> Tuple[bool,str,List]:
        """
    Check if the joint configuration q is in the free configuration space, considering
    joint limits, collision with obstacles and auto collision. It also outputs a message about a
    possible violation.

    For efficiency purposes, the program halts in the first violation it finds (if there is any).
    So, the message, if any, is only about one of the possible violations. There can be more.

    Parameters
    ----------
    q : a nD vector (n-element list/tuple, (n,1)/(1,n)/(n,)-shaped numpy matrix/numpy array)
        The manipulator's joint configuration.
        (default: the current joint configuration (robot.q) for the manipulator).

    obstacles : list of simple objects (see Utils.IS_SIMPLE)
        A list of obstacles as simple objects 
        (default: empty list).

    htm : 4x4 numpy matrix
        The robot base's configuration.
        (default: the same as the current htm).

    check_joint: boolean
        If joint limits should be considered or not.
        (default: True).

    check_auto: boolean
        If auto collision should be considered or not.
        (default: True).

    tol : positive float
        Tolerance for convergence in the iterative algorithm, in meters.
        (default: 0.0005 m).

    dist_tol : positive float
        The tolerance to consider that two links are colliding.
        (default: 0.005 m).

    no_iter_max : positive int
        The maximum number of iterations for the distance computing algorithm.
        (default: 20 iterations).

    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto').

    Returns
    -------
    is_free : boolean
        If the configuration is free or not.

    message: string
        A message about what is colliding (is otherwise just 'Ok!').

    info: list
        Contains a list with information of the violation of free space (if there is any, otherwise
        it is empty). The first element is the violation type: either 0 (lower joint limit violated), 1 (upper joint
        limit violated) 2 (collision with obstacles) and 3 (collision between links). The last elements depends on
        the violation type.

        If lower joint limit or upper joint limit, contains which joint was violated.

        If collision with obstacles, contains the list [i,isub,j], containing the index of the link i
        i, the index of the collision object in the link isub and the obstacle index in the list, j.

        If collision, contains the list [i,isub,j,jsub], containing the index of the first link i,
        the index of the collision object in the first link isub, the index of the second link j and
        the index of the collision object in the second link jsub.


    """

        return _check_free_config(self, q, htm, obstacles, check_joint, check_auto, tol, dist_tol, no_iter_max, mode)



     
    #######################################################################################################   
    #LEGACY
    def check_free_configuration(self, q: Optional[Vector]=None, htm: Optional[HTMatrix]=None, 
                          obstacles: List[MetricObject]=[], check_joint: bool = True, check_auto: bool = True,
                          tol: float = 0.0005, dist_tol: float = 0.005, no_iter_max: int = 20, 
                          mode: str = 'auto') -> Tuple[bool,str,List]:
      
      
        # Backward compatibility shim
        import warnings
        warnings.warn(
            "'check_free_configuration' is deprecated, use 'check_free_config' instead.",
            DeprecationWarning
        )
              
        return _check_free_config(self, q, htm, obstacles,
                              check_joint, check_auto,
                              tol, dist_tol, no_iter_max, mode)
     
    #######################################################################################################



import uaibot as ub
import numpy as np
import requests
from io import BytesIO
import matplotlib.pyplot as plt
import sys
import os



#sim = ub.Demo.constrained_control_demo_1()


# Get the point cloud data
url = "https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/PointCloud/data_wall_with_hole.npy"
wallpoints = np.load(BytesIO(requests.get(url).content))
pc = ub.PointCloud(points = wallpoints, size=0.02, color='cyan')


#Create the other objects and set up the simulation
point_link = ub.Ball(color='red',radius=0.02)
point_obj = ub.Ball(color='blue',radius=0.02)
robot = ub.Robot.create_franka_emika_3()
htm_des = ub.Utils.trn([0.45,0.31,0.7])*ub.Utils.rotx(-np.pi/2)
frame_des = ub.Frame(htm_des, size=0.1)
sim = ub.Simulation.create_sim_grid([pc, robot, point_link, point_obj, frame_des])
sim.set_parameters(width=500,height=500,load_screen_color='#191919',background_color='#191919')
d_safe_jl = (np.pi/180)*5
robot.add_ani_frame(0,[0.0, 0.0, 0.0, -np.pi*4/180-2*d_safe_jl, 0.0, 2*d_safe_jl, 0.0])

#Main loop
t = 0
dt= 0.01
cont = True

min_dobs = 1e6
min_dauto = 1e6
min_djoint = 1e6
for i in range(1000):
    
    #Get current configuration
    qr = robot.q
    
    #Get smooth distance structure from robot to point cloud
    qdot, ep, eo, dobs, dauto, djoint =robot.constrained_control(htm_tg=htm_des, q=qr, obstacles=pc)

    #Integrate joint velocity
    qr += qdot*dt
    
    min_dobs = min(min_dobs,dobs)
    min_dauto = min(min_dauto,dauto)
    min_djoint = min(min_djoint,djoint)
    
  

    
    sys.stdout.write("\rep = "+str(round(1000*ep,2))+", eo = "+str(round(eo,2))+", dobs = "+str(round(min_dobs,4))+", dauto = "+str(round(min_dauto,4))+", djoint = "+str(round(min_djoint,4)))
    sys.stdout.flush()
    
    #Update animation frames and time
    robot.add_ani_frame(t,qr)        
    t+=dt
    
    
current_folder = os.path.dirname(os.path.abspath(__file__))
sim.save(current_folder,"control")
